/*
**
** Tcl Extension for Windows
** RCS Version $Revision: 1.7 $
** RCS Last Change Date: $Date: 2002/10/07 18:03:11 $
** Original Author: Michael I. Schwartz, mschwart@nyx.net
** 
** {LICENSE}
** 
** THIS SOFTWARE IS PROVIDED BY THE AUTHOR WITH PERMISSION TO USE, COPY, MODIFY, AND
** DISTRIBUTE IT FOR ANY PURPOSE WITH THE FOLLOWING CONDITIONS:
** 1) IN SOURCE FORM, THIS HEADER MUST BE PRESERVED AND THESE CONDITIONS PROPOGATED
** 2) IN BINARY FORM, THE ORIGINAL AUTHOR MUST BE ACKNOWLEDGED IN DOCUMENTATION AND
**    CREDITS SCREEN (IF ANY)
** 3) FOR ANY COMMERCIAL SALE, THE AUTHOR MUST BE NOTIFIED OF THE USE OF THIS CODE
**    IN A PARTICULAR PRODUCT; THE ORIGINAL AUTHOR MAY CITE 
**    THE COMMERCIAL SELLER AS A BENEFICIARY OF THE CODE.
** 
** 
** IN NO EVENT SHALL THE AUTHOR BE LIABLE TO ANY PARTY FOR DIRECT,
** INDIRECT, SPECIAL, INCIDENTAL,  OR CONSEQUENTIAL DAMAGES ARISING OUT OF
** THE USE OF THIS SOFTWARE, ITS DOCUMENTATION,  OR ANY DERIVATIVES
** THEREOF, EVEN IF THE AUTHORS HAVE BEEN ADVISED OF THE POSSIBILITY OF
** SUCH DAMAGE.
** 
** THE AUTHORS AND DISTRIBUTORS SPECIFICALLY DISCLAIM ANY WARRANTIES,
** INCLUDING, BUT NOT LIMITED TO,  THE IMPLIED WARRANTIES OF
** MERCHANTABILITY,FITNESS FOR A PARTICULAR PURPOSE, AND NON-INFRINGEMENT. 
** THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, AND THE AUTHORS AND
** DISTRIBUTORS HAVE NO OBLIGATION  TO PROVIDE MAINTENANCE, SUPPORT,
** UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
** 
** {SYNOPSIS}
** 
** This file contains commands to extend TK for Windows 3.11, Windows 95,
** and Windows NT 4.0 features
** 
** The commands are:
** 
** ini
** 
** The details of each command's options follow:
** 
** {PROFILE}
** 
** ini [get|set|delete] ...
** ini get [-section sectname] [-private filename] \
**                     [-default defstring] [-key string] [-max maxsize]
**  DESCRIPTION:
**    Used to get information from Win3.1 .INI files.
**    These are supported by Win95 and WinNT as well, but the registry is strongly 
**    preferred by these systems.
**  OPTIONS:
**    -private 
**      filename points to a particular .INI file. Default is WIN.INI (or equivalent).
**    -section 
**      sectname points to the proper section of the .INI file. 
**      If omitted, a list ofall sections will be returned.
**    -default 
**      defstring is a string to return if the key is not found 
**      The default is an empty string ("")
**    -key 
**      string is the value to search for. 
**      If omitted, a list of keys in the section is returned
**    -max maxsize is the biggest return value expected (default 1024).
**      for example, the call ini get -section fonts would normally 
**      require a bigger -max.
**  LIMITATIONS:
**    Under Win32s (Win3.11), failure to provide a section will NOT list all sections.
** 
** ini set [-section sectname] [-private filename] -key keyname -value valuename
**  DESCRIPTION:
**    Similarly, sets values in the .INI files.
** 
** ini delete -section sectname -private filename [-key keyname]
**  DESCRIPTION:
**    Similarly, deletes either the value for the key or the entire section 
**    (if no key is provided)
** 
** ini version
**    Returns the version of this package
****************************************************************==*/

#if defined(__WIN32__) || defined (__WIN32S__) || defined (WIN32S)
  #   if defined(_MSC_VER)
  #     define EXPORT(a,b) __declspec(dllexport) a b
  #     define DllEntryPoint DllMain
  #   else
  #     if defined(__BORLANDC__)
  #         define EXPORT(a,b) a _export b
  #     else
  #         define EXPORT(a,b) a b
  #     endif
  #   endif
#else
  #   error "This extension is only for Windows. Use tkuxini.c for other OS!"
#endif

#include <windows.h>
#include <tcl.h>
#include <string.h>
#include <stdlib.h>

/* New macros for tcl8.0.3 and later */
#if defined(TCL_STORAGE_CLASS)
#  undef TCL_STORAGE_CLASS
#endif
#define TCL_STORAGE_CLASS DLLEXPORT
#if ! defined(EXTERN)
#  define EXTERN
#endif

#if TCL_MAJOR_VERSION == 7 && TCL_MINOR_VERSION <= 5
/* In this case, must replace Tcl_Alloc(), Tcl_Realloc(), and Tcl_Free()
** with things based in Tcl_Ckalloc() and Tcl_Ckfree()
*/

#define Tcl_Alloc(x)  Tcl_Ckalloc(x)
#define Tcl_Free(x)   Tcl_Ckfree(x)

/* Hmm. Emulate this one by always reallocating and copying all the bytes.
** It may not be efficient, but ought to do the job
*/
char *		Tcl_Realloc (char *ptr, unsigned int size)
{
  char *newptr = Tcl_Alloc(size);
  memcpy(newptr, ptr, size);
  Tcl_Ckfree(ptr);
  return newptr;
}
#endif

/* Exported symbols */
BOOL APIENTRY DllEntryPoint (HINSTANCE hInstance, DWORD reason, LPVOID lpCmdLine)
{
  /* Don't do anything, so just return true */
  return TRUE;
}

/****************************************************************
** External function prototypes
****************************************************************/
EXTERN int Ini_Init (Tcl_Interp *interp);
EXTERN int Ini_SafeInit (Tcl_Interp *interp);

/****************************************************************
** Internal function prototypes
****************************************************************/
static int DoCommand ( Tcl_Interp *interp, int argc, char **argv, int safe);
static int Version(Tcl_Interp *interp);
static int Ini (ClientData unused, Tcl_Interp *interp, int argc, char **argv);
static int GetIni (ClientData unused, Tcl_Interp *interp, int argc, char **argv);
static int SetIni (ClientData unused, Tcl_Interp *interp, int argc, char **argv);
static int DelIni (ClientData unused, Tcl_Interp *interp, int argc, char **argv);

/****************************************************************
** Internal static data structures (ClientData)
****************************************************************/
static char safe_usage_message[] = "Usage: ini [get|version]";
static char unsafe_usage_message[] = "Usage: ini [get|set|delete|version]";
static char *default_usage_message = unsafe_usage_message;
static char working_buffer[1024+1];

/****************************************************************
** Support for getting error messages out of windows error codes
****************************************************************/
static struct ErrorMap
{
  long errnum;
  char *msg;
} WinErrorMap[] =
{
  { 0,                            "\"Unexpected error code\"" },
  { ERROR_FILE_NOT_FOUND,         "\"Requested value or subkeys of nonexistent key\"" },
  { ERROR_ACCESS_DENIED,          "\"Cannot delete registry key in NT with subkeys\"" },
  { ERROR_BADDB,                  "\"Registry database is corrupt\"" },
  { ERROR_BADKEY,                 "\"Configuration registry key is invalid\"" },
  { ERROR_CANTOPEN,               "\"Can't open registry key\"" },
  { ERROR_CANTREAD,               "\"Can't read registry key\"" },
  { ERROR_CANTWRITE,              "\"Can't write registry key\"" },
  { ERROR_REGISTRY_RECOVERED,     "\"Successfully recovered a corrupt registry file\"" },
  { ERROR_REGISTRY_CORRUPT,       "\"Registry database is corrupt and unrecoverable\"" },
  { ERROR_REGISTRY_IO_FAILED,     "\"Registry I/O operattion failed\"" },
  { ERROR_NOT_REGISTRY_FILE,      "\"Registry attempted to load non-registry file\"" },
  { ERROR_KEY_DELETED,            "\"Operation cannot be completed on registry key marked for deletion\""},
  { ERROR_NO_LOG_SPACE,           "\"Could not allocate required space in Registry log\""},
  { ERROR_KEY_HAS_CHILDREN,       "\"Cannot create symbolic link in registry key with subkeys or values\"" },
  { ERROR_CHILD_MUST_BE_VOLATILE, "\"Cannot create stable subkey under volatile parent key\"" },
};

static int FormatWinError(long code, char *buffer, int maxsize)
{
  int i;
  char numbuf[13]; /* long enough for the biggest long number */
  int len;
  sprintf(numbuf, "%ld ", code);
  len = strlen(numbuf);
  if (maxsize <= len)
  {
    return -1;
  }
  strcpy(buffer, numbuf);
  buffer += len;
  maxsize -= len;
  
  for (i=1; i< sizeof(WinErrorMap) / sizeof(struct ErrorMap); i++)
  {
    if ( code == WinErrorMap[i].errnum )
    {
      strncpy(buffer, WinErrorMap[i].msg, maxsize);
      buffer[maxsize - 1] = '\0';
      return 0;
    }
  }
  strncpy (buffer, WinErrorMap[0].msg, maxsize);
  buffer[maxsize - 1] = '\0';
  return 1;
} 

/****************************************************************
** Changed from version 0.2 to make ini a key word, followed
** by either get, set, or delete
****************************************************************/
static int Ini (ClientData unused, Tcl_Interp *interp, int argc, char **argv)
{
  if ( argc == 0 )
  {
    Tcl_SetResult(interp, default_usage_message, TCL_STATIC);
    return TCL_ERROR;
  }

  if ( strcmp(argv[0], "get") == 0 )
  {
    argc--;
    argv++;
    return GetIni(unused, interp, argc, argv);
  }
  else if ( strcmp(argv[0], "set") == 0 )
  {
    argc--;
    argv++;
    return SetIni(unused, interp, argc, argv);
  }
  else if ( strcmp(argv[0], "delete") == 0)
  {
    argc--;
    argv++;
    return DelIni(unused, interp, argc, argv);
  }
  else if ( strcmp(argv[0], "version") == 0)
  {
    return Version(interp);
  }
  
  Tcl_SetResult(interp, default_usage_message, TCL_STATIC);
  return TCL_ERROR;
}

/****************************************************************
** ini set -section sectname [-private filename ] -key key -value value
****************************************************************/
static int SetIni (ClientData unused, Tcl_Interp *interp, int argc, char **argv)
{
  int i;
  BOOL status;
  char *inifile = 0;
  char *key = 0;
  char *value = 0;
  char *sectionname   = 0;
  char *illegal       = 0;
  static char usage_message[] = "Usage: ini set -section sectname "
                                "[-private filename] -key key -value value";
  
  for (i=0; i<argc; i++)
  {
    if ( strcmp (argv[i], "-private") == 0 )
      inifile = argv[++i];
    else if ( strcmp (argv[i], "-section") == 0)
      sectionname = argv[++i];
    else if ( strcmp (argv[i], "-key") == 0)
      key = argv[++i];
    else if ( strcmp (argv[i], "-value") == 0)
      value = argv[++i];
    else
      illegal = argv[i];
  }

  if ( sectionname == 0 || key == 0 || value == 0 || illegal != 0 ) /* Error: no key, value, or section */
  {
    Tcl_SetResult(interp, usage_message, TCL_STATIC);
    return TCL_ERROR;
  }

  if (inifile)
  {
    status = WritePrivateProfileString (sectionname, key, value, inifile);
  }
  else
  {
    status = WriteProfileString (sectionname, key, value);
  }
  
  if (status == TRUE)
    return TCL_OK;

  FormatWinError(GetLastError(), working_buffer, sizeof(working_buffer));
  Tcl_SetResult(interp, working_buffer, TCL_VOLATILE);
  return TCL_ERROR;
}

/****************************************************************
** ini delete -section sectname [-private filename ] [-key key]
****************************************************************/
static int DelIni (ClientData unused, Tcl_Interp *interp, int argc, char **argv)
{
  int i;
  BOOL status;
  char *inifile = 0;
  char *key = 0;
  char *value = 0;
  char *sectionname   = 0;
  char *illegal       = 0;
  static char usage_message[] = "Usage: ini delete -section sectname "
                                "[-private filename] [-key key]";
  
  for (i=0; i<argc; i++)
  {
    if ( strcmp (argv[i], "-private") == 0 )
      inifile = argv[++i];
    else if ( strcmp (argv[i], "-section") == 0)
      sectionname = argv[++i];
    else if ( strcmp (argv[i], "-key") == 0)
      key = argv[++i];
    else
      illegal = argv[i];
  }

  if ( sectionname == 0 || illegal != 0 ) /* Error: no key, value, or section */
  {
    Tcl_SetResult(interp, usage_message, TCL_STATIC);
    return TCL_ERROR;
  }

  if (inifile)
  {
    status = WritePrivateProfileString (sectionname, key, value, inifile);
  }
  else
  {
    status = WriteProfileString (sectionname, key, value);
  }
  
  if (status == TRUE)
    return TCL_OK;

  FormatWinError(GetLastError(), working_buffer, sizeof(working_buffer));
  Tcl_SetResult(interp, working_buffer, TCL_VOLATILE);
  return TCL_ERROR;
}

/****************************************************************
**  ini get [-section sectname] [-private filename ] [-default default] [-key string] [-max size]
****************************************************************/
static int GetIni (ClientData unused, Tcl_Interp *interp, int argc, char **argv)
{
  int i;
  DWORD status;
  char *inifile = 0;
  char *inistring = 0;
  char *defaultstring = "";
  char *sectionname   = 0;
  char *buffer;
  DWORD bufsize = 1024;
  char *illegal = 0;
  static char usage_message[] = "Usage: ini get [-section sectname] "
                                "[-private filename] "
				"[-default defaultstring] "
				"[-key string] [-max maxsize]";
  
  for (i=0; i<argc; i++)
  {
    if ( strcmp (argv[i], "-private") == 0 )
      inifile = argv[++i];
    else if ( strcmp (argv[i], "-section") == 0)
      sectionname = argv[++i];
    else if ( strcmp (argv[i], "-default") == 0)
      defaultstring = argv[++i];
    else if ( strcmp (argv[i], "-key") == 0)
      inistring = argv[++i];
    else if ( strcmp (argv[i], "-max") == 0)
    {
      DWORD newsize;
      char *str = argv[++i];
      newsize = str?atol(str):0;
      if (newsize > bufsize)
        bufsize = newsize;
    }
    else
      illegal = argv[i];
  }

  if ( defaultstring == 0 || illegal != 0 ) /* Error: no section */
  {
    Tcl_SetResult (interp, usage_message, TCL_STATIC);
    return TCL_ERROR;
  }

  if ( (buffer = Tcl_Alloc(bufsize*sizeof(char))) == 0 )
  {
    Tcl_SetResult(interp,
                  "Out of memory in get_ini. Try setting -max larger",
                  TCL_STATIC);
    return TCL_ERROR;
  }
  memset(buffer, '\0', bufsize); /* 0 out memory for use */

  if (inifile)
    status = GetPrivateProfileString(sectionname, inistring, defaultstring,
                                     buffer, bufsize, inifile);
  else
    status = GetProfileString(sectionname, inistring, defaultstring,
                                     buffer, bufsize);
  buffer[bufsize - 1] = '\0'; /* Ensure result is null terminated */
  
  if ( sectionname == 0 || inistring == 0 ) /* Got all ini strings */
  {
    char *cp = buffer;
    while (*cp)
    {
      Tcl_AppendElement(interp, cp);
      cp += strlen(cp) + 1;
    }
  }
  else /* Just one ini string */
  {
    Tcl_SetResult(interp, buffer, TCL_VOLATILE);
  }
  
  Tcl_Free(buffer);
  
  return TCL_OK;
}

/****************************************************************
** Ini
** This is the "unsafe" version which supports the full range
** of possibilities.
****************************************************************/
static int ini (ClientData data, Tcl_Interp *interp, int argc, char **argv)
{
  if (argc > 1)
  {
    if ( strcmp (*argv, "ini") == 0 )
      return Ini(data, interp, argc-1, argv+1);
  }
  Tcl_SetResult(interp, unsafe_usage_message, TCL_STATIC);
  return TCL_ERROR;
}

/****************************************************************
** This is the "safe" version of tkmswin, which supports only
** printing and the version command
****************************************************************/
static int ini_safe (ClientData data, Tcl_Interp *interp, 
                         int argc, char **argv)
{
  if (argc > 1)
  {
    if ( strcmp (*argv, "ini") == 0 )
    {
      /* Only support get and version */
      argv++;
      argc--;
      if ( strcmp (*argv, "get") == 0 )
      {
        argc--;
        argv++;
        return GetIni(data, interp, argc, argv);
      }
      else if  ( strcmp (*argv, "version") == 0 )
        return Version(interp);
    }
  }
  Tcl_SetResult(interp, safe_usage_message, TCL_STATIC);
  return TCL_ERROR;
}

/****************************************************************
** This is the version information/command
** The static data should also be used by pkg_provide, etc.
****************************************************************/
/* Version information */
static char version_string[] = "1.0.0.0";

/* Version command */
static int Version(Tcl_Interp *interp)
{
  Tcl_SetResult (interp, version_string, TCL_STATIC);
  return TCL_OK;
}

/****************************************************************
** Initialization procedures
** These are the only public procedures in the file.
****************************************************************/
/* Initialization Procedures */
EXTERN int Ini_Init (Tcl_Interp *interp)
{
  #if defined(USE_TCL_STUBS)
    Tcl_InitStubs(interp, TCL_VERSION, 0 );
  #endif
  
#if TCL_MAJOR_VERSION <= 7
  Tcl_CreateCommand(interp, "ini", ini, 0, 0);
#else
  /* Wanted to use namespaces, but "unknown" isn't smart enough yet */
  Tcl_CreateCommand(interp, "ini", ini, 0, 0);
#endif
  Tcl_PkgProvide (interp, "ini", version_string);
    
  default_usage_message = unsafe_usage_message;

  return TCL_OK;
}

EXTERN int Ini_SafeInit (Tcl_Interp *interp)
{
  #if defined(USE_TCL_STUBS)
    Tcl_InitStubs(interp, TCL_VERSION, 0 );
  #endif

  Tcl_CreateCommand(interp, "ini", ini_safe, 0, 0);
  Tcl_PkgProvide (interp, "ini", version_string);

  default_usage_message = safe_usage_message;

  return TCL_OK;
}
